/**
 * http://www.stat.ubc.ca/~bouchard/GEP/README-usage.html
 * 
 * @author Alexandre Bouchard
 * @author Ardavan Saeedi
 */

package gep;
import static nuts.util.CollUtils.list;

import java.io.File;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.Set;

import nuts.io.CSV;
import nuts.io.IO;
import nuts.lang.StringUtils;
import nuts.util.CollUtils;
import nuts.util.Indexer;
import pty.smc.ParticleFilter;
import fig.basic.IOUtils;
import fig.basic.LogInfo;
import fig.basic.Option;
import fig.basic.StrUtils;
import fig.exec.Execution;
import fig.record.PathRecordNode;
import gep.comparisons.EMForCTMC;
import gep.comparisons.FlatReconstructionMethod;
import gep.comparisons.HeldoutReconstructionMethod;
import gep.comparisons.IncrementalReconstructionMethod;
import gep.data.DataSource;
import gep.data.GeneratedData;
import gep.data.LoadFlatDataFormat;
import gep.data.MSData;
import gep.data.RnaDataset;
import gep.model.SplitContext;
import gep.pmcmc.PMCMC;
import gep.pmcmc.PartialHiddenState;
import gep.timeseries.Measurements;
import gep.timeseries.Series;
import gep.util.OutputManager;


public class GEPMain implements Runnable
{
  @Option public int maxNSeries = 100;
  @Option public DataType dataType = DataType.GENERATED;
  
  @Option public double heldOutPr = 0.1;
  @Option public Random heldoutRandom = new Random(1);
  @Option public Random samplingRand = new Random(1);
  
  @Option public int maxMCMCIters = 2500;
  
  @Option public boolean saveReconstruction = false;
  
  @Option public Method method = Method.SHGEP;
  
  @Option public String pathToHeldoutSpec = ""; // empty string meas they are generated using the specified heldOutPr
  
  public static enum Method { SHGEP, EM }
  
  public static OutputManager outputManager = new OutputManager();
  public static Indexer<String> observedIndexer = null;//Encodings.rnaEncodings().nonGapCharactersIndexer();
  public static GeneratedData generatedData = null;
  public static RnaDataset rnaData = null;
  private static MSData msData;
  private static LoadFlatDataFormat loadPlainTxtData;
  public static enum DataType { GENERATED, RNA, MS, LOAD_PLAIN_TXT }
  
  public static int nChars() { return observedIndexer.size(); }
  
  private static PMCMC pmcmc;
  private static EMForCTMC em;
  private List<Measurements> allAnswers = list();
  private List<Measurements> rawObservations = list();
  
  private IncrementalReconstructionMethod system;
  
  
  public static void main(String[] args)
  {
    pmcmc = new PMCMC();
    em = new EMForCTMC();
    PMCMC.pf = new ParticleFilter<PartialHiddenState<SplitContext>>();
    generatedData = new GeneratedData();
    rnaData = new RnaDataset();
    msData = new MSData();
    loadPlainTxtData = new LoadFlatDataFormat();
    IO.run(args, new GEPMain(), "pmcmc", pmcmc, "em", em, "pf", PMCMC.pf, "gen", generatedData, "rna", rnaData, "msdata", msData, "paintxt", loadPlainTxtData);
  }
  
  @Override
  public void run()
  {
    // main method
    
    
    if (method == Method.SHGEP)
      system = pmcmc;
    else if (method == Method.EM)
      system = em;
    else
      throw new RuntimeException();
    
    File reconstructionsDir = new File(Execution.getFile("reconstructions"));
    reconstructionsDir.mkdir();
    
//    observedIndexer = dataSource().indexer();
    List<String> observations = list(dataSource().possibleObservations());
    Collections.sort(observations); // make the indexing canonical (otherwise runs with same seed will have different outcomes)
    observedIndexer = new Indexer<String>(observations);
    // init observation, goal stats
    initData();
    // compute baseline
    PrintWriter out = null;
    if (saveReconstruction) out = IOUtils.openOutHard(new File(reconstructionsDir, "FLAT_MODEL"));
    final double baseLineError = reconstructionError(out, FlatReconstructionMethod.mleFlatRecon(rawObservations), allAnswers);
    if (saveReconstruction) out.close();
    outputManager.printWrite("errors", "Method", "FLAT_MODEL", "Iter", 0, "Error", baseLineError);
    // init hidden states
    system.init(samplingRand);
    // do some sampling
//    LogInfo.logs("Sampling seed:" + samplingRand.nextLong());
    for (int mcmcIter = 0; mcmcIter < maxMCMCIters; mcmcIter++)
    {
      LogInfo.track("Iteration " + mcmcIter + "/" + maxMCMCIters, true);
      
      system.iterate(samplingRand, mcmcIter);
      
      if (saveReconstruction) out = IOUtils.openOutHard(new File(reconstructionsDir, "" + method + "-" + mcmcIter));
      final double curError = reconstructionError(out, system, allAnswers);
      if (saveReconstruction) out.close();
      outputManager.printWrite("errors", "Method", method, "Iter", mcmcIter, "Error", curError);
      
      LogInfo.end_track();
    }
    outputManager.close();
  }
  
  
  public static double reconstructionError(PrintWriter out, HeldoutReconstructionMethod method, List<Measurements> allAnswers)
  {
    
    if (out != null)
      out.println(CSV.header("seriesIndex", "queryPointIndex", "truthIndex", "reconstructedIndex"));
    
    double nEval = 0.0;
    double nErrors = 0.0;
    for (int sIdx = 0; sIdx < allAnswers.size(); sIdx++)
    {
      Measurements curAns = allAnswers.get(sIdx);
      nEval += curAns.size();
      for (int qIdx = 0; qIdx < curAns.size(); qIdx++)
      {
        final int answer = curAns.getValue(qIdx);
        final int guess = method.reconstruct(sIdx, qIdx);
        if (answer != guess)
          nErrors++;
        
        if (out != null)
          out.println(CSV.body(sIdx, qIdx, answer, guess));
      }
    }
    return nErrors/nEval;
  }

  // returns: observation data, answers to queries
  private void initData()
  {
    loadHelout();
    
    PrintWriter output = openOut(),
                heldoutOutput = heldoutOpenOut();
    int nSeq = 0;
    int nDat = 0;
    int nHeld = 0;
    List<Series> allObservations = list();
    for (int i = 0; i < maxNSeries; i++)
    {
      final Measurements currentMeasurement = dataSource().next(observedIndexer);
//      System.out.println("read meas " + i + ":" + currentMeasurement);
      if (currentMeasurement == null)
        break;
      // will hold the observations
      List<Double> times = list(); 
      List<Integer> values = list();
      // will hold the query times and answers
      List<Double> queryTimes = list();
      List<Integer> answers = list();
      
      holdout(currentMeasurement, times, values, queryTimes, answers, i);
      saveData(output, heldoutOutput, i, currentMeasurement, queryTimes);
      
      nSeq++;
      nDat += currentMeasurement.size();
      nHeld += queryTimes.size();
      
      Measurements currentO = new Measurements(times, values);
      rawObservations.add(currentO);
      allObservations.add(new Series(currentO, queryTimes));
      Measurements currentQ = new Measurements(queryTimes, answers);
      allAnswers.add(currentQ);
    }
    system.loadData(allObservations, nChars());
    LogInfo.logsForce("nSequencesLoaded: " + nSeq);
    LogInfo.logsForce("nDatapointsLoaded: " + nDat);
    LogInfo.logsForce("nPointsHeldout: " + nHeld);
    output.close();
    heldoutOutput.close();
  }
  
  public boolean usingLoadedHeldoutSpec()
  {
    return pathToHeldoutSpec != null && !pathToHeldoutSpec.equals("");
  }
  
  private List<Set<Double>> loadedHeldoutTimes = null;
  private void loadHelout()
  {
    if (usingLoadedHeldoutSpec())
    {
      LogInfo.logs("Holding out using loaded file:" + pathToHeldoutSpec);
      loadedHeldoutTimes = list();
      int index = 0;
      for (String line : IO.i(pathToHeldoutSpec))
        if (!LoadFlatDataFormat.IGNORE_LINE_MATCH.matcher(line).matches())
        {
          String [] mainSplit = line.split("[:]");
          if (mainSplit.length != 1 && mainSplit.length != 2)
            throw new RuntimeException();
          int specifiedSeqIdx = Integer.parseInt(StringUtils.selectFirstRegex("Sequence\\s+([0-9]+)$", mainSplit[0]));
          if (specifiedSeqIdx != index)
            throw new RuntimeException();
          String [] subSplit = (mainSplit.length > 1 ? mainSplit[1].split("\\s+") : new String[0]);
          Set<Double> currentSet = CollUtils.set();
          for (String field : subSplit)
            currentSet.add(Double.parseDouble(field));
          loadedHeldoutTimes.add(currentSet);
          index++;
        }
    }
    else
      LogInfo.logs("Holding out with probability " + heldOutPr + " and using random seed " + heldoutRandom.nextLong());
  }

  private void saveData(PrintWriter output,  PrintWriter heldoutOutput, int i, Measurements currentMeasurement,
      List<Double> queryTimes)
  {
    output.println("# Sequence " + i );
    output.println(StrUtils.join(currentMeasurement.getTimes(), " "));
    for (int value : currentMeasurement.getValues())
      output.append(observedIndexer.i2o(value) + " ");
    output.println();
//    output.println(StrUtils.join(currentMeasurement.getValues(), " "));
    heldoutOutput.println("Sequence " + i + ":" + StrUtils.join(queryTimes, " "));
  }

  private PrintWriter openOut()
  {
    PrintWriter out = IOUtils.openOutHard(Execution.getFile("data.txt"));
    out.println("# Time series represented by space-separated line doublets:");
    out.println("#   Line 1: The times of the measurements");
    out.println("#   Line 2: The states at these times (any string that does not contain space/tabs)");
    out.println("# Line starting by # or empty lines are ignored");
    out.println();
    return out;
  }
  
  private PrintWriter heldoutOpenOut()
  {
    PrintWriter out = IOUtils.openOutHard(Execution.getFile("heldout.txt"));
    out.println("# Heldout points for each series:");
    out.println("#   Format for each line: Sequence [n]: a space separated list of times, exactly as shown in the series file (NOT indices) heldout for sequence [n]");
    out.println("# Line starting by # or empty lines are ignored");
    out.println("");
    return out;
  }

  private DataSource dataSource()
  {
    if (dataType == DataType.GENERATED)
      return generatedData;
    else if (dataType == DataType.RNA)
      return rnaData;
    else if (dataType == DataType.MS)
      return msData;
    else if (dataType == DataType.LOAD_PLAIN_TXT)
      return loadPlainTxtData;
    else
      throw new RuntimeException();
  }

  private void holdout(
      Measurements allMeasurements,
      List<Double> times,
      List<Integer> values, 
      List<Double> queryTimes, 
      List<Integer> answers,
      int seqIdx)
  {
    // make sure the heldout time are subsets
    Set<Double> allTimes = CollUtils.set(allMeasurements.getTimes());
    if (usingLoadedHeldoutSpec() && !allTimes.containsAll(loadedHeldoutTimes.get(seqIdx)))
      throw new RuntimeException();
    // do the job
    for (int currentSqIdx = 0; currentSqIdx < allMeasurements.size(); currentSqIdx++)
    {
      final int currentObsState = allMeasurements.getValue(currentSqIdx);
      final double currentTime = allMeasurements.getTime(currentSqIdx); 
      if (shouldHoldOut(currentSqIdx, allMeasurements.size(), currentTime,  seqIdx)) // do not hold out the last one to avoid problem
      {
        // holding out!
        queryTimes.add(currentTime);
        answers.add(currentObsState);
      }
      else
      {
        // observed!
        times.add(currentTime);
        values.add(currentObsState);
      }
    }
  }
  
  private boolean shouldHoldOut(int currentSqIdx, int nMeas, double currentTime, int seqIdx)
  {
    boolean isLast = currentSqIdx == nMeas - 1;
    if (usingLoadedHeldoutSpec())
    {
      boolean isInSet = loadedHeldoutTimes.get(seqIdx).contains(currentTime);
      if (isLast && isInSet)
      {
        LogInfo.warning("Currently, holding out the last item in the sequence is not supported for implementation reasons.  Ignoring held out time " + currentTime + " in sequence index " + seqIdx);
        return false;
      }
      return isInSet;
    }
    else
      return heldoutRandom.nextDouble() < heldOutPr  &&  !isLast;
  }
}
