/**
 * http://www.stat.ubc.ca/~bouchard/GEP/README-usage.html
 * 
 * @author Alexandre Bouchard
 * @author Ardavan Saeedi
 */
package gep.pmcmc;
import static nuts.util.CollUtils.list;
import static nuts.util.MathUtils.checkClose;

import java.util.Collections;
import java.util.List;
import java.util.Random;

import nuts.math.MeasureZeroException;
import nuts.util.CollUtils;
import nuts.util.Counter;
import nuts.util.CounterMap;

import org.apache.commons.math.stat.descriptive.SummaryStatistics;

import fig.basic.LogInfo;
import fig.basic.Option;
import gep.GEPMain;
import gep.comparisons.IncrementalReconstructionMethod;
import gep.model.Predictives;
import gep.model.SplitContext;
import gep.model.SplitHyperParams;
import gep.model.SufficientStatisticsImpl;
import gep.timeseries.Series;

import pty.smc.ParticleFilter;
import pty.smc.ParticleFilter.StoreProcessor;


public class PMCMC implements IncrementalReconstructionMethod
{
  @Option public double initBaseMeasureAlpha0 = 1;
  @Option public double initAlphaL1 = 1; // these can be resampled easily (their post is gamma)
  @Option public double initAlphaL2 = 1;
  @Option public double initAlphaL3 = 1;
  @Option public double initBeta0 = 1;
  @Option public double initGamma0 = 1;
  
  @Option public boolean generateHittingTimes = false;
  
  @Option public boolean sampleGamma = false;
  
  @Option public boolean useIncrementalInit = true;
  @Option public int nInitRetries = 10;
  
  
  public static ParticleFilter<PartialHiddenState<SplitContext>> pf = null;
  private  List<Series> series;
  private  Predictives<SplitContext,SplitContext> predictiveDistributions;
  public SummaryStatistics acceptanceRate = null,
                           totalPropFailure=null;
  
  @Override
  public int reconstruct(int sIdx, int qIdx)
  {
    CounterMap<Integer,Integer> post = series.get(sIdx).getQueryPosteriors();
    Counter<Integer> curPost = post.getCounter(qIdx);
    checkClose(curPost.totalCount(), 1.0);
    return curPost.argMax();
  }
  
  @Override
  public void init(Random samplingRand)
  {
    initSequence(samplingRand, nInitRetries, useIncrementalInit);
  }

  @Override
  public void iterate(Random samplingRand, int iterIndex)
  {
    resetAcceptRateStats();
//    for (int sIdx = 0; sIdx < nSeries(); sIdx++)
    List<Integer> indices = CollUtils.ints(nSeries());
    Collections.shuffle(indices, samplingRand);
    for (int sIdx : indices)
    {
      sampleSequence(samplingRand, sIdx);
      if (generateHittingTimes)
        Generator.generateHitting(predictiveDistributions.readOnlySuffStats, predictiveDistributions, samplingRand, 5000, 10000, SplitContext.BEG);
      if (iterIndex > 0 && sampleGamma)
        sampleGammas(samplingRand);
    }
    
    LogInfo.logsForce("gammas:" + getGammas());
    GEPMain.outputManager.printWrite("totalFailRate", "Iter", iterIndex, "FailRate", totalPropFailure.getMean());
    GEPMain.outputManager.printWrite("acceptRate", "Iter", iterIndex, "AcceptRate", acceptanceRate.getMean());
    GEPMain.outputManager.printWrite("nHiddenStates", "Iter", iterIndex, "NumHiddenStates", nHiddenStates());
    
    updateAllQueryStats();
  }
  
  @Override
  public void loadData(List<Series> series, int nChars)
  {
    this.predictiveDistributions = 
      Predictives.splitPredictives(
          new SufficientStatisticsImpl<SplitContext,SplitContext>(), 
          initBaseMeasureAlpha0, 
          initAlphaL1, 
          initAlphaL2, 
          initAlphaL3, 
          nChars, 
          initBeta0,
          initGamma0);
    this.series = series;
  }
  
  public int nHiddenStates()
  {
    return SufficientStatisticsImpl.nHiddenStates(predictiveDistributions.readOnlySuffStats);
  }
  
  public void updateAllQueryStats()
  {
    for (Series s : series)
      s.updateQueryStats();
  }
  
  public int nSeries() { return series.size(); }

  public void initSequence(Random rand, int nInitRetries, boolean useIncrementalInit)
  {
    for (int seqnIndex = 0; seqnIndex < series.size(); seqnIndex++)
    {
      boolean accepted = false;
      for (int retry = 0; retry<nInitRetries+1 && !accepted; retry++)
      {
        final int multiple = (int) Math.pow(2, retry);
        pf.N = pf.N * multiple;
        try     
        { 
          accepted = sampleSequence(rand, seqnIndex, useIncrementalInit); 
        }
        finally 
        { 
          pf.N = pf.N / multiple; 
        }
      }
      if (!accepted)
        throw new RuntimeException("Failed to initialize sequence");
    }
    // commit all in one shot
    if (!useIncrementalInit)
      for (int i = 0; i < series.size(); i++)
        SufficientStatisticsImpl.plusEqual(predictiveDistributions.readOnlySuffStats, series.get(i).suffStats);
  }
  public boolean sampleSequence(Random rand, int seqnIndex)
  {
    return sampleSequence(rand, seqnIndex, true);
  }
  public boolean sampleSequence(Random rand, int seqnIndex, boolean commitSuffStats)
  {
    boolean accept = false;
    Series current = series.get(seqnIndex);
    
    
    // remove current from global
    if (commitSuffStats)
      SufficientStatisticsImpl.minusEqualCheckNonNeg(predictiveDistributions.readOnlySuffStats, current.suffStats);
    // compute PF using SMCKernel
    SMCKernel<SplitContext> kernel = new SMCKernel<SplitContext>(predictiveDistributions, current.observations, SplitContext.BEG);
    StoreProcessor<PartialHiddenState<SplitContext>> processor = 
      new StoreProcessor<PartialHiddenState<SplitContext>>();
    double ratio = 0.0;
    try
    {
      pf.sample(kernel, processor);
      // sample one particle
      PartialHiddenState<SplitContext> sampled = processor.sample(rand);
      // compute new logLL estimate
      final double newLogLL = pf.estimateNormalizer();
      // compute accept
      ratio = Math.min(1.0, Math.exp(newLogLL - current.previousLogLL));
      accept = rand.nextDouble() < ratio;
//      LogInfo.logs("Accepted:" + accept);
      if (accept)
      {
//        LogInfo.logs(Generator.toString(sampled.eventsList(), new Indexer<Double>(), GEPMain.observedIndexer));
        // if accepted, update series.suffStat
        current.suffStats = sampled.suffStats();
        // if accepted, update previousLogLL
        current.previousLogLL = newLogLL;
        // if accepted, update currentHiddenVariables
        current.currentHiddenVariables = sampled.eventsList();
        current.initialized = true;
      }
      if (totalPropFailure != null)
        totalPropFailure.addValue(0.0);
    } catch (MeasureZeroException mze)
    {
      if (totalPropFailure != null)
        totalPropFailure.addValue(1.0);
    }
    finally
    {
      if (acceptanceRate != null)
        acceptanceRate.addValue(ratio);
      // add series.suffStat back to global
      if (commitSuffStats)
        SufficientStatisticsImpl.plusEqual(predictiveDistributions.readOnlySuffStats, current.suffStats);
    }
    return accept;
  }
  
  public void sampleGammas(Random rand)
  {
    ((SplitHyperParams)predictiveDistributions.hyperParams).sample(predictiveDistributions, rand);
  }
  
  public List<Double> getGammas()
  {
    return ((SplitHyperParams) predictiveDistributions.hyperParams).getGammas();
  }

  public void resetAcceptRateStats()
  {
    totalPropFailure = new SummaryStatistics();
    acceptanceRate = new SummaryStatistics();
  }
}
