/**
 * http://www.stat.ubc.ca/~bouchard/GEP/README-usage.html
 * 
 * @author Alexandre Bouchard
 * @author Ardavan Saeedi
 */
package gep.pmcmc;
import static nuts.util.CollUtils.list;
import static nuts.util.MathUtils.close;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import nuts.io.IO;
import nuts.util.Indexer;
import pepper.Encodings;
import pty.smc.ParticleFilter;
import pty.smc.ParticleKernel;
import fig.basic.Pair;
import fig.basic.StrUtils;
import gep.model.Predictives;
import gep.model.SplitContext;
import gep.model.SufficientStatistics;
import gep.model.SufficientStatisticsImpl;
import gep.timeseries.Event;
import gep.timeseries.Measurements;


public final class SMCKernel<S> implements ParticleKernel<PartialHiddenState<S>>
{
  private final Predictives<S,S> predictiveDistributions;
  private final Measurements observations;
  private final S begSymbol;
  
  public SMCKernel(Predictives<S, S> predictiveDistributions,
      Measurements obs, S begSymbol)
  {
    this.predictiveDistributions = predictiveDistributions;
    this.observations = obs;
    this.begSymbol = begSymbol;
  }

  @Override
  public PartialHiddenState<S> getInitial()
  {
    return new PartialHiddenState<S>(new SufficientStatisticsImpl<S,S>(), new ArrayList<Event<S>>(), 0);
  }

  @Override
  public int nIterationsLeft(PartialHiddenState<S> partialState)
  {
    return nObservations() - partialState.nObservationAskedToCover();
  }
  
  public double observationTimes(int i)
  {
    return observations.getTime(i);
  }
  public int observation(int i)
  {
    return observations.getValue(i);
  }
  public int nObservations() { return observations.size(); }

  /**
   * returns null if there are no next state possible (e.g. because of a
   * structured base measure)
   * @param rand
   * @param current
   * @return (state,LOG weight update) of the next state sampled
   */
  @Override
  public Pair<PartialHiddenState<S>, Double> next(
      Random rand,
      PartialHiddenState<S> current)
  {
    
    final int obsIndexToCover = current.nObservationAskedToCover();
    final double timeOfObservationToAdd = observationTimes(obsIndexToCover);
    final List<Event<S>> eventsList = list(current.eventsList());
    double delta = timeOfObservationToAdd - current.eventsLength();
    if (delta < 0.0) // this means the observation is already covered (and checks have been made)
    {
      SufficientStatistics<S,S> newSS = new SufficientStatisticsImpl<S,S>(current.suffStats());
      return Pair.makePair(new PartialHiddenState<S>(newSS, eventsList, current.nObservationAskedToCover()+1), 0.0);
    }
    SufficientStatistics<S,S> newSS = 
      Generator.generate(current.suffStats(), predictiveDistributions, rand, eventsList, delta, Integer.MAX_VALUE, begSymbol);
    
    // check if new event is consistents with observations (current and possibly next ones)
    // if inconsistent, just return weight of -INFINITY
    double logW = 0.0; 
    
    
    // only the last event can cross new time indices
    if (eventsList.size() < 2)
      throw new RuntimeException(); //"size="+eventsList.size() + "\n" + eventsList); 
    //   ERROR: Caused by java.lang.RuntimeException: size=1:
    Event<S> e1 = eventsList.get(eventsList.size() - 2),
             e2 = eventsList.get(eventsList.size() - 1);
    S sc = Event.contents(e1, e2);
    double right= Event.right(e1, e2);
    
    loop:for (int i = obsIndexToCover; i < nObservations(); i++)
    {
      if (observationTimes(i) > right)
      {
        if (i == obsIndexToCover) 
          throw new RuntimeException(); //"\n" + StrUtils.join(Generator.lastLines,"\n") + "\n");
        break loop;
      }
      else if (observation(i) != ((SplitContext) sc).getX())
      {
        logW = Double.NEGATIVE_INFINITY;
        break loop;
      }
    }
    return Pair.makePair(new PartialHiddenState<S>(newSS, eventsList, current.nObservationAskedToCover()+1), logW);
  }
  
  // some tests
//  public static void main(String [] args)
//  {
//    int nParts = 100;
//    Random rand = new Random(1);
//    Indexer<Character> observedIndexer = Encodings.rnaEncodings().nonGapCharactersIndexer();
////    Indexer<Double> hiddenIndexer = new Indexer<Double>();
//    double bmAlpha0 = 100, alphaL1 = 100, alphaL2 = 100, alphaL3 = 100, beta0 = 100, gamma0 = 100;
//    Predictives<SplitContext,SplitContext> pred = 
//      Predictives.splitPredictives(
//          new SufficientStatisticsImpl<SplitContext,SplitContext>(), 
//          bmAlpha0, 
//          alphaL1, 
//          alphaL2, 
//          alphaL3, 
//          4, 
//          beta0, 
//          gamma0);
//    // basic test: check if sampled states match with observations:
//    Measurements obs = randomObservations(rand, 4, 1.0, 10);
//    System.out.println(obs.toString(observedIndexer) + "\n\n");
//    SMCKernel<SplitContext> kernel = new SMCKernel<SplitContext>(pred, obs, SplitContext.BEG);
//    ParticleFilter.bootstrapFilter(kernel, new SplitContextParticleProcessor(), nParts, rand);
//  }
//  
//  public static class SplitContextParticleProcessor implements ParticleFilter.ParticleProcessor<PartialHiddenState<SplitContext>>
//  {
//    Indexer<Character> observedIndexer = Encodings.rnaEncodings().nonGapCharactersIndexer();
//    Indexer<Double> hiddenIndexer = new Indexer<Double>();
//    @Override
//    public void process(PartialHiddenState<SplitContext> state, double weight)
//    {
//      IO.so("Weight:" + weight);
//      if (!close(0.0, weight))
//        IO.so(Generator.toString(state.eventsList(), hiddenIndexer, observedIndexer));
//      IO.so();
//    }
//  }
  
  public static Measurements randomObservations(Random rand, int nSymb, double delta, int nObs)
  {
    List<Double> times = list();
    List<Integer> values = list();
    double cTime = 0;
    for (int i = 0; i < nObs; i++)
    {
      times.add(cTime);
      values.add(rand.nextInt(nSymb));
      cTime += delta;
    }
    return new Measurements(times, values);
  }

}
